import time

def size_to_human(size):
    if (size >= 1073741824):
        return "%.1fG" % (size / 1073741824)
    if (size > 1024 * 1024):
        return "%.1fM" % (size / 1048576)
    if (size > 1024):
        return "%.1fK" % (size / 1024)
    return str(int(size))

def time_to_human(t):
    return time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(t/1000))

def format_right_text(text, padding=0):
    return ("{:>"+str(padding)+"}").format(text)

def format_time(t, padding=0):
    return format_right_text(time_to_human(t), padding)

def format_human_size(size, padding=0):
    return format_right_text(size_to_human(size), padding)

def _format_size(size, padding=0):
    return format_right_text(str(size), padding)

def _format_text(text, padding=0):
    return text

class Table:

    def __init__(self, *formats):
        self.formats = formats
        self.rows = []

    def append(self, *row):
        self.rows.append(row)

    def print(self):
        formatters = [self._formatter_lookup(f) for f in self.formats]
        sizes = [0 for _ in self.formats]

        for row in self.rows:
            for (ix, formatter) in enumerate(formatters):
                if formatter:
                    cell = formatter(row[ix], 0)
                    sizes[ix] = max(sizes[ix], len(cell))

        for row in self.rows:
            cells = []
            for (ix, formatter) in enumerate(formatters):
                if formatter:
                    cells.append(formatter(row[ix], sizes[ix]))
            print(" ".join(cells))

    def _formatter_lookup(self, format):
        if format is None:
            return None
        if format == "size":
            return _format_size
        if format == "human_size":
            return format_human_size
        if format == "text":
            return _format_text
        if format == "right_text":
            return format_right_text
        if format == "time":
            return format_time
        raise Exception(f"No formatter available for type {format}")



import pathlib
import subprocess
import json
import logging

def mkdirs(path):
    return pathlib.Path(path).mkdir(parents=True, exist_ok=True)

def call_external_processor_json(cmd, data):
    cp = subprocess.run(cmd,
                        shell=True,
                        input = (json.dumps(data)+"\n").encode("utf-8") ,
                        capture_output=True)
    logging.debug(f"external processor output: {cp.stdout}")
    return json.loads(cp.stdout.decode("utf-8"))



import sys
import asyncio
import aiohttp
import os
import os.path
import posixpath
import urllib.parse
import time
import base64
import traceback
import base64
import json
import logging
import inspect

from fastdbfs.fileinfo import FileInfo
from fastdbfs.swarm import Swarm
from fastdbfs.exceptions import RateError, APIError
import fastdbfs.util

class FindEntry():
    def __init__(self, fi, good=True, ex=None, done=False):
        self.fi = fi
        self.good = good
        self.ex = ex
        self.done = done

    def __str__(self):
        return f"FindEntry(path={self.fi.abspath()}, done={self.done}, ex={self.ex})"

    def __repr__(self):
        return self.__str__()

class Disconnected():
    def __init__(self):
        pass

    def __getattr__(self, method):
        raise Exception("open must be called first!")

    def prompt(self):
        return "*disconnected* "

class DBFS():

    def __init__(self, id, host, cluster_id, token,
                 workers=8, chunk_size=1048576, max_retries=10,
                 error_delay=10, error_delay_increment=10):
        self.id = id
        self.host = host
        self.cluster_id = cluster_id
        self.token = token
        self.cwd = "/"
        self.workers = workers
        self.chunk_size = chunk_size
        self.max_retries = max_retries
        self.error_delay = error_delay
        self.error_delay_increment = error_delay_increment
        self.wait_until = 0

        self.semaphore = asyncio.Semaphore(workers)

        # print(f"host: {self.host}, cluster: {self.cluster_id}, token: {self.token}")

    def check(self):
        self._assert_dir(".")

    def _resolve(self, path, *more):
        return posixpath.normpath(posixpath.join(self.cwd, path, *more))

    def prompt(self):
        if self.id == "DEFAULT":
            return f"{self.cwd}$ "
        else:
            return f"{self.id}:{self.cwd}$ "

    def cd(self, path):
        path = self._resolve(path)
        self._assert_dir(path)
        self.cwd = path

    def _asynchronize(self, cb):
        if cb is None:
            return None
        async def async_cb(*args, **kwargs):
            return cb(*args, **kwargs)
        return async_cb

    def _make_swarm(self, name=None, queue_max_size=None, **kwargs):
        if name is None:
            name = inspect.stack()[1].function
        if queue_max_size is None:
            queue_max_size=self.workers
        return Swarm(self.workers, queue_max_size=queue_max_size, name=name, **kwargs)

    async def _unpack_http_response(self, response):
        status = response.status

        if response.headers.get("x-envoy-ratelimited", None) == "true":
            # print(f"Too many requests!")
            raise RateError("Too many requests")

        if "content-type" not in response.headers:
            print(f"Response headers: {response.headers}")

        ct = response.headers['content-type']
        if ct != "application/json":
            raise Exception(f"Unexpected response received from API, Content-type: {ct}, HTTP status: {status}")

        try:
            data = await response.json()
        except OSError as ex:
            # connection errors and alike should be retried
            raise ex
        except Exception as ex:
            logging.debug("Unable to retrieve API response {ex}", exc_info=True)
            raise Exception("Unable to retrieve API response") from ex

        if status == 200:
            return data

        try:
            error_code = data["error_code"]
            message = data["message"]
        except:
            raise Exception("Unexpected response received from API, HTTP status {status}")

        raise APIError(error_code, message)

    async def _async_control(self, cb):
        retries = 0
        while True:
            async with self.semaphore:
                try:
                    delta = self.wait_until - time.time()
                    if delta > 0:
                        # print(f"Delaying for {delta} seconds")
                        await asyncio.sleep(delta)
                    return await cb()
                except RateError as ex:
                    self.wait_until = time.time() + 1
                    # we do not increase the retries count when a
                    # RateError pops up!
                except OSError as ex:
                    if retries > self.max_retries:
                        raise ex
                    # We keep the semaphore while we sleep in order to
                    # reduce the pressure on the remote side.
                    await asyncio.sleep(self.error_delay + self.error_delay_increment * retries)
                    retries += 1

                except Exception as ex:
                    logging.debug("Stack trace:", exc_info=True)
                    raise ex

    def _async_http_get(self, session, end_point, **params):
        async def cb():
            url = urllib.parse.urljoin(self.host, end_point)
            headers = { "Authorization": "Bearer " + self.token }
            #print(f"session: {session}\nurl: {url}\nparams: {params}\nheaders: {headers}")
            async with session.get(url,
                                   params=params,
                                   headers=headers) as response:
                return await self._unpack_http_response(response)
        return self._async_control(cb)

    def _async_http_post(self, session, end_point, **data):
        async def cb():
            url = urllib.parse.urljoin(self.host, end_point)
            headers = { "Authorization": "Bearer " + self.token,
                        "Content-Type": "application/json" }
            load = json.dumps(data)
            async with session.post(url,
                                    headers=headers,
                                    data=load) as response:
                return await self._unpack_http_response(response)

        return self._async_control(cb)

    async def _async_call_with_session(self, cb, *params, **kwparams):
        async with aiohttp.ClientSession() as session:
            return await cb(session, *params, **kwparams)

    async def _async_get_status(self, session, path):
        r = await self._async_http_get(session, "api/2.0/dbfs/get-status",
                                       path=self._resolve(path))
        return FileInfo.from_json(r)

    def _run_with_session(self, task, *args, **kwargs):
        task_with_session = self._async_call_with_session(task, *args, **kwargs)
        loop = asyncio.get_event_loop()
        return loop.run_until_complete(task_with_session)

    def _simple_get(self, end_point, **params):
        return self._run_with_session(self._async_http_get, end_point, **params)

    def _simple_post(self, end_point, **params):
        return self._run_with_session(self._async_http_post, end_point, **params)

    def get_status(self, path):
        return self._run_with_session(self._async_get_status, path)

    def _async_mkdir(self, session, path):
        path = self._resolve(path)
        return self._async_http_post(session, "api/2.0/dbfs/mkdirs", path=path)

    def mkdir(self, path):
        return self._run_with_session(self._async_mkdir, path=path)

    async def _async_list(self, session, path):
        path = self._resolve(path)
        #print(f"> {path}")
        r = await self._async_http_get(session, "api/2.0/dbfs/list", path=path)
        #print(f"listing {path} done")
        return [FileInfo.from_json(e) for e in r.get("files", [])]

    def ls(self, path):
        path = self._resolve(path)
        fi = self.get_status(path)
        if fi.is_dir():
            return self._run_with_session(self._async_list, path=path)
        else:
            return [fi]

    def _async_rm(self, session, path, recursive=False):
        path = self._resolve(path)
        return self._async_http_post(session, "api/2.0/dbfs/delete",
                                     path=path, recursive=recursive)

    def rm(self, path, recursive=False):
        path = self._resolve(path)
        self._run_with_session(self._async_rm,
                               path=path, recursive=recursive)

        # Have we removed the CWD?
        if self.cwd != "/":
            if (self.cwd == path or
                (self.cwd.startswith(path) and self.cwd[len(path)] == "/")):
                # Yes, fix it!
                self.cwd = posixpath.split(path)[0]

    def _assert_dir(self, path):
        if not self.get_status(path).is_dir():
            raise Exception("Not a directory")

    async def _async_create(self, session, path, overwrite=False):
        path = self._resolve(path)
        out = await self._async_http_post(session,
                                     "api/2.0/dbfs/create",
                                     path=path, overwrite=overwrite)
        return int(out["handle"])

    def _async_add_block(self, session, handle, block):
        return self._async_http_post(session,
                                     "api/2.0/dbfs/add-block",
                                     data=base64.standard_b64encode(block).decode('ascii'),
                                     handle=handle)

    def _async_close(self, session, handle):
        return self._async_http_post(session,
                                     "api/2.0/dbfs/close",
                                     handle=handle)

    async def _async_put_from_file(self, session, infile, target, size=None, overwrite=False,
                                   update_cb=None):
        target = self._resolve(target)

        handle = await self._async_create(session, target, overwrite)
        try:
            bytes_copied = 0
            while True:
                chunk = infile.read(self.chunk_size)
                chunk_len = len(chunk)
                if chunk_len == 0:
                    break

                await self._async_add_block(session, handle, chunk)

                bytes_copied += chunk_len
                if update_cb:
                    await update_cb(size=size, bytes_copied=bytes_copied)

            await self._async_close(session, handle)

            fi = await self._async_get_status(session, target)
            if fi.size() != bytes_copied:
                raise Exception(f"corrupted copy detected, copied: {bytes_copied}, remote: {fi.size()}, expected: {size}")
            return fi
        except Exception as ex:
            try: await self._async_rm(target)
            except: pass
            raise ex

    def put_from_file(self, infile, target, size=None, overwrite=False, update_cb=None):
            return self._run_with_session(self._async_put_from_file,
                                          infile, target, size, overwrite,
                                          self._asynchronize(update_cb))

    async def _async_put(self, session, src, target, **kwargs):
        size = os.stat(src).st_size
        with open(src, "rb") as infile:
            return await self._async_put_from_file(session, infile, target, size, **kwargs)

    def put(self, src, target, overwrite, update_cb):
        return self._run_with_session(self._async_put,
                                      src, target,
                                      overwrite=overwrite,
                                      update_cb=self._asynchronize(update_cb))

    async def _async_get_chunk(self, session, path, offset, length, out):
        # print(f"processing chunk at {offset}")
        remaining = length
        while remaining > 0:
            r = await self._async_http_get(session,
                                           "api/2.0/dbfs/read",
                                           path=path, offset=offset, length=length)

            bytes_read = r["bytes_read"]
            if bytes_read == 0 or bytes_read > remaining:
                raise Exception("Invalid data response")

            chunk = base64.standard_b64decode(r["data"])
            if len(chunk) != bytes_read:
                raise Exception("Invalid data response, size is not as promised")

            out.seek(offset)
            out.write(chunk)
            remaining -= bytes_read
            offset += bytes_read
        # print(f"chunk of {length} bytes copied at offset {offset - length}")
        return length

    async def _async_get(self, session, low_swarm, src, target,
                         overwrite=False, mkdirs=False, update_cb=None):

        if mkdirs:
            fastdbfs.util.mkdirs(os.path.dirname(target))

        # FIXME: a race condition exists here!
        if not overwrite and os.path.exists(target):
            raise Exception("File already exists")
        with open(target, "wb") as out:
            await self._async_get_to_file(session, low_swarm, src, out, update_cb)

    async def _async_get_to_file(self, session, low_swarm, src, out, update_cb=None):
        src = self._resolve(src)
        fi = await self._async_get_status(session, src)
        size = fi.size()

        response_queue = asyncio.Queue()
        active_tasks = 0 # we count the request we have queued
        bytes_copied = 0
        offset = 0

        try:
            while True:
                while True:
                    try:
                        res = response_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    active_tasks -= 1
                    bytes_copied += Swarm.unwrap_response(res)
                    if update_cb:
                        await update_cb(size, bytes_copied)

                next_offset = min(offset + self.chunk_size, size)
                length = next_offset - offset
                if length <= 0:
                    break

                await low_swarm.put(self._async_get_chunk,
                                    response_queue=response_queue,
                                    path=src, offset=offset, length=length,
                                    out=out)
                active_tasks += 1
                offset = next_offset

            while active_tasks > 0:
                res = await response_queue.get()
                active_tasks -= 1
                bytes_copied += Swarm.unwrap_response(res)
                if update_cb:
                    await update_cb(size, bytes_copied)

        finally:
            for _ in range(active_tasks):
                try: await response_queue.get()
                except: pass

        return fi

    def get_to_file(self, src, out, update_cb=None):
        swarm = self._make_swarm()
        task =  self._async_call_with_session(self._async_get_to_file,
                                              swarm, src, out,
                                              self._asynchronize(update_cb))
        return swarm.loop_until_complete(task)

    async def _async_find(self, session, path, update_cb, filter_cb=None, predicates={}):
        path = self._resolve(path)

        async def filter(entries):
            entries_by_relpath = { e.fi.relpath(path): e for e in entries }
            for relpath, e in entries_by_relpath.items():
                if not e.fi.check_predicates(relpath=relpath, **predicates):
                    e.good = False
            if filter_cb:
                input = { k: v.fi for k, v in entries_by_relpath.items() }
                selected_paths = await filter_cb(input)
                for p in selected_paths:
                    try:
                        del entries_by_relpath[p]
                    except KeyError:
                        logging.warning(f"Ignoring unexpected path {p} returned by filter")
                for e in entries_by_relpath.values():
                    e.good = False

        root_fi = await self._async_get_status(session, path)
        root_entry = FindEntry(root_fi)
        await filter([root_entry])

        if not root_fi.is_dir():
            await update_cb(entry=root_entry, entries_found=1, done=1)
            return

        swarm = self._make_swarm(use_priority_queue=True, queue_max_size=0)
        response_queue = asyncio.PriorityQueue()

        async def control():
            await swarm.put(self._async_list,
                            task_key = path,
                            response_queue = response_queue,
                            path = path)
            pending = [root_entry]
            entries_found = 1
            while pending:
                logging.debug(f"_async_find control() loop, first pending entry is {pending[0].fi.abspath()} of a total of {len(pending)}")
                (key, fis, ex) = await response_queue.get()
                logging.debug(f"_async_find control() loop, got response for {key}");
                for e in pending:
                    if e.fi.abspath() == key:
                        e.ex = ex
                        e.done = True
                        break
                else:
                    raise Exception("Internal error, key not found in pending")

                if fis:
                    entries_found += len(fis)
                    new = [FindEntry(fi, done=not fi.is_dir())
                           for fi in fis]
                    await filter(new)
                    pending += new
                    pending.sort(key=lambda x: x.fi.abspath())

                    for fi in fis:
                        if fi.is_dir():
                            await swarm.put(self._async_list,
                                            task_key = fi.abspath(),
                                            response_queue = response_queue,
                                            path = fi.abspath())

                # finally report entries ready from the top
                while pending and pending[0].done:
                    e = pending.pop(0)
                    logging.debug(f"_async_back control() passing up {e.fi.abspath}, good: {e.good}")
                    await update_cb(entry=e, max_entries=entries_found, done=entries_found-len(pending))
            logging.debug("_async_find control() done")

        await swarm.run_while(control())

    def find(self, path, update_cb, filter_cb=None, predicates={}):
        self._run_with_session(self._async_find,
                               self._resolve(path),
                               self._asynchronize(update_cb),
                               self._asynchronize(filter_cb),
                               predicates=predicates)

    async def _async_rget(self, session, src, target, update_cb=None, filter_cb=None, predicates={}):
        src = self._resolve(src)

        # We use two worker queues here, low is for the low level
        # requests. We pass it to the _self_get method without ever
        # using it directly. high_queue is for parallelizing the
        # _self_get calls.

        low_swarm = self._make_swarm(name="rget-low")
        high_swarm = self._make_swarm(name="rget-high")
        response_queue = asyncio.Queue()

        entries = 1 # those are the entries seen so far as reported by find
        active_gets = {}

        async def update(entry):
            if update_cb:
                await update_cb(entry=entry, max_entries=entries, done=entries-len(active_gets))

        async def update_after_get(relpath, ex):
            entry = active_gets.pop(relpath)
            entry.ex = ex
            logging.debug("Exception caught during file get")
            await update(entry)

        async def empty_response_queue():
            try:
                while True:
                    (relpath, _, ex) = response_queue.get_nowait()
                    await update_after_get(relpath, ex)
            except asyncio.QueueEmpty:
                return

        async def find_cb(entry, max_entries, done):
            nonlocal entries

            # We empty the response_queue everytime this function is called.
            await empty_response_queue()

            fi = entry.fi
            relpath = fi.relpath(src)
            local_path = os.path.abspath(os.path.join(target, relpath))
            entries = max_entries
            if fi.is_dir():
                if entry.good: # mkdir
                    logging.debug(f"making dir for {relpath}")
                    try:
                        fastdbfs.util.mkdirs(local_path)
                    except Exception as ex:
                        if not os.path.isdir(local_path):
                            logging.warn(f"{relpath}: mkdirs failed. {ex}")
                            entry.ex = ex
                else:
                    logging.debug(f"discarding dir {relpath}")
            else:
                if entry.good:
                    logging.debug(f"queueing download of {relpath}")
                    await high_swarm.put(self._async_get,
                                         task_key=relpath,
                                         response_queue=response_queue,
                                         low_swarm=low_swarm,
                                         src=fi.abspath(), target=local_path,
                                         mkdirs=True)
                    active_gets[relpath] = entry
                    return # don't report the entry yet!
                else:
                    logging.debug(f"discarding file {relpath}")

            await update(entry)

        # This code is a bit hairy because we run the control code and
        # both swarms concurrently:
        async def control():
            await high_swarm.run_while(self._async_find(session, src,
                                                        update_cb=find_cb,
                                                        filter_cb=filter_cb,
                                                        predicates=predicates))

            while active_gets:
                logging.debug(f"_async_rget control, {len(active_gets)} gets currently active")
                (relpath, _, ex) = await response_queue.get()
                await update_after_get(relpath, ex)

        await low_swarm.run_while(control())

    def rget(self, src, target, update_cb=None, filter_cb=None, predicates={}):
        self._run_with_session(self._async_rget,
                               src, target,
                               filter_cb=self._asynchronize(filter_cb),
                               update_cb=self._asynchronize(update_cb),
                               predicates=predicates)

    async def _async_rput(self, session, src, target, overwrite=False, update_cb=None):
        target = self._resolve(target)
        local_root_fi = FileInfo.from_local(src)

        swarm = self._make_swarm()
        response_queue = asyncio.Queue()

        async def control():
            active_puts = 0
            for (dir, _, filenames) in os.walk(src):
                local_fi = FileInfo.from_local(dir)
                relpath = local_fi.relpath(local_root_fi.abspath())
                remote_path = self._resolve(target, relpath)
                local_path = os.path.join(src, relpath)

                status = None
                try:
                    await self._async_mkdir(session, path=remote_path)
                    if update_cb:
                        await update_cb(local_path, True, None)

                except Exception as ex:
                    if update_cb:
                        await update_cb(local_path, False, ex)
                        for fn in filenames:
                            await update_cb(os.path.join(local_path, fn), False, None)

                else:
                    for fn in filenames:
                        local_child_path = os.path.join(local_path, fn)
                        remote_child_path = self._resolve(remote_path, fn)
                        #print(f"remote_child_path: {remote_child_path}")
                        await swarm.put(self._async_put,
                                        task_key=local_child_path,
                                        src=local_child_path,
                                        target=remote_child_path,
                                        overwrite=overwrite,
                                        response_queue=response_queue)
                        active_puts += 1

                while True:
                    try:
                        (local_child_path, res, ex) = response_queue.get_nowait()
                    except asyncio.QueueEmpty:
                        break
                    active_puts -= 1
                    if update_cb:
                        ok = ex is None
                        await update_cb(local_child_path, ok, ex)

            while active_puts > 0:
                (local_child_path, res, ex) = await response_queue.get()
                active_puts -= 1
                if update_cb:
                    ok = ex is None
                    await update_cb(local_child_path, ok, ex)

        await swarm.run_while(control())

    def rput(self, src, target, overwrite=False, update_cb=None):
        self._run_with_session(self._async_rput, src, target,
                               overwrite=overwrite,
                               update_cb=self._asynchronize(update_cb))




import stat

import os
import os.path
import posixpath
import logging

class FileInfo():

    @staticmethod
    def from_json(json):
        return FileInfo(json["is_dir"],
                        json["file_size"],
                        json["modification_time"],
                        json["path"])

    @staticmethod
    def from_local(filename, *more):
        path = os.path.abspath(os.path.join(filename, *more))
        st = os.stat(path)
        return FileInfo(stat.S_ISDIR(st.st_mode),
                        st.st_size,
                        st.st_mtime,
                        path)

    def __init__(self, is_dir, size, mtime, abspath):
        self._is_dir = is_dir
        self._size = size
        self._mtime = mtime
        self._abspath = abspath

    def to_data(self):
        return { "abspath": self._abspath,
                 "size":    self._size,
                 "mtime":   self._mtime,
                 "is_dir":  self._is_dir }

    def basename(self):
        return posixpath.basename(self._abspath)

    def abspath(self):
        return self._abspath

    def relpath(self, base, requested=None):
        if requested is not None and requested[0] == "/":
            # the original request was absolute
            return self._abspath
        base = posixpath.normpath(base)
        path = self.abspath()
        if path == base:
            return "."
        if path.startswith(base):
            if base == "/":
                return path[1:]
            elif path[len(base)] == "/":
                return path[(len(base)+1):]
        return path

    def is_dir(self):
        return self._is_dir

    def mtime(self):
        return self._mtime

    def size(self):
        return self._size

    def type(self):
        return "dir" if self._is_dir else "file"

    def _check_predicate__newer_than(self, limit, _):
        return self.mtime() >= limit*1000

    def _check_predicate__older_than(self, limit, _):
        return self.mtime() <= limit*1000

    def _check_predicate__max_size(self, limit, _):
        if self.is_dir():
            return True
        return self.size() <= limit

    def _check_predicate__min_size(self, limit, _):
        if self.is_dir():
            return True
        return self.size() >= limit

    # FIXME: all those methods doing exactly the same are ugly
    def _check_predicate__name(self, pattern, _):
        return pattern.fullmatch(self.basename(), _)

    def _check_predicate__iname(self, pattern):
        return pattern.fullmatch(self.basename(), _)

    def _check_predicate__re(self, pattern):
        return pattern.search(self.basename(), _)

    def _check_predicate__ire(self, pattern, _):
        return pattern.search(self.basename())

    def _check_predicate__iwholere(self, pattern, relpath):
        return pattern.search(self.basename if relpath is None else relpath)

    def _check_predicate__wholere(self, pattern, relpath):
        return pattern.search(self.basename if relpath is None else relpath)

    def check_predicates(self, relpath=None, **predicates):
        #print(f"predicates: {predicates}")
        for key, value in predicates.items():
            if value is not None:
                if key.startswith("exclude_"):
                    method = getattr(self, f"_check_predicate__{key[len('exclude_'):]}")
                    if method(value, relpath):
                        logging.debug(f"entry {self.abspath()} has been discarded by predicate {key} with value {value}")
                        return False
                else:
                    method = getattr(self, f"_check_predicate__{key}")
                    if not method(value, relpath):
                        logging.debug(f"entry {self.abspath()} has been discarded by predicate {key} with value {value}")
                        return False
        return True



import sys
import cmd
import configparser
import os
import os.path
import posixpath
import traceback
import time
import shlex
import pathlib
import subprocess
import tempfile
import logging
import progressbar

from fastdbfs.dbfs import DBFS, Disconnected
from fastdbfs.cmdline import option, flag, arg, remote, local, argless, chain
from fastdbfs.format import Table, format_human_size, format_time
import fastdbfs.util

_find_predicates=chain(option("min-size", cast="size"),
                       option("max-size", cast="size"),
                       option("max-depth", cast="int"),
                       option("min-depth", cast="int"),
                       option("iname", cast="glob", cast_args={"case_insensitive": True}),
                       option("name", cast="glob"),
                       option("re", "regexp", cast="re"),
                       option("ire", "iregexp", cast="re", cast_args={"case_insensitive": True}),
                       option("wholere", "whole-regexp", cast="re"),
                       option("iwholere", "iwhole-regexp",
                              cast="re", cast_args={"case_insensitive": True}),
                       option("newer-than", "newer", cast="date>"),
                       option("older-than", "older", cast="date<"),
                       option("exclude-iname", cast="glob", cast_args={"case_insensitive": True}),
                       option("exclude-name", cast="glob"),
                       option("exclude-re", "exclude-regexp", cast="re"),
                       option("exclure-ire", "exclude-iregexp",
                              cast="re", cast_args={"case_insensitive": True}),
                       option("exclude-wholere", "exclude-whole-regexp", cast="re"),
                       option("exclude-iwholere", "exclude-iwhole-regexp",
                              cast="re", cast_args={"case_insensitive": True}),
                       option("external-filter", "ext-filter"))

class CLI(cmd.Cmd):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._init_cfg()
        self._dbfs = Disconnected()
        self._set_prompt()
        self._debug = True

    def cfg(self, section, key, default=None):
        try:
            return self._cfg[section][key]
        except:
            if default is not None:
                return default
            raise Exception(f"Configuration entry {key} missing in section {section}")

    def cfg_int(self, section, key):
        v = self.cfg(section, key)
        try:
            return int(v)
        except:
            raise Exception(f"Configuration entry {key} is not an integer")

    def _init_cfg(self):
        self._cfg = configparser.ConfigParser()
        self._cfg["fastdbfs"] = {
            "workers": 8,
            "chunk_size": 1048576,
            "max_retries": 10,
            "pager": "less",
            "error_delay": 10,
            "error_delay_increment": 10,
            "log_level": "WARNING"
        }
        self._cfg["logging"] = {}

        fns = [os.path.expanduser(fn) for fn in ("~/.databrickscfg", "~/.fastdbfs", "~/.config/fastdbfs")]
        self._cfg.read(fns)

    def _do_open(self, profile):
        dbfs = DBFS(profile,
                    host = self.cfg(profile, "host"),
                    cluster_id = self.cfg(profile, "cluster_id"),
                    token = self.cfg(profile, "token"),
                    chunk_size = self.cfg_int("fastdbfs", "chunk_size"),
                    workers = self.cfg_int("fastdbfs", "workers"),
                    max_retries = self.cfg_int("fastdbfs", "max_retries"))

        dbfs.check()
        self._dbfs = dbfs

    @arg("profile", default="DEFAULT")
    def do_open(self, profile):
        """
        open [profile]

        Sets the active Databricks profile.
        """

        self._do_open(profile)

    @remote("path")
    def do_cd(self, path):
        """
        cd [path]

        Change the remote current directory.
        """

        self._dbfs.cd(path)

    def do_mkcd(self, arg):
        """
        mkcd [path]

        Creates the given directory and then sets it as the current directory.
        """

        self.do_mkdir(arg)
        self.do_cd(arg)

    @remote("path")
    def do_mkdir(self, path):
        """
        mkdir path

        Creates the remote path.
        """

        self._dbfs.mkdir(path)

    def _do_ls(self, path, long, human):
        """
        ls [OPTS] [path]

        List the contents of the remote directory.

        The accepted options are as follows:

          -l, --long   Print file properties.

          -h, --human  Print file sizes in a human friendly manner.
        """

        if long:
            size_format = "human_size" if human else "size"
            table = Table("right_text", size_format, "time", "text")
        else:
            table = Table(None, None, None, "text")
        for fi in self._dbfs.ls(path):
            table.append(fi.type(), fi.size(), fi.mtime(), fi.basename())
        table.print()

    @flag("long", "l")
    @flag("human", "h")
    @remote("path", default=".")
    def do_ls(self, path, long, human):
        return self._do_ls(path, long, human)

    @flag("human", "h")
    @remote("path", default=".")
    def do_ll(self, path, human):
        """
        ll [OPTS] [path]

        ll is an alias for "ls -l".
        """
        return self._do_ls(path, True, human)

    @local("path", default="~")
    def do_lcd(self, path):
        """
        lcd [path]

        Changes the local working directory.
        """

        os.chdir(path)

    @argless()
    def do_lpwd(self):
        """
        lpwd

        Displays the local current directory.
        """

        print(os.getcwd())

    @flag("overwrite", "o")
    @local("src")
    @remote("target", arity="?")
    def do_put(self, overwrite, src, target):
        """
        put [OPTS] src [target]

        Copies the given local file to the remote system.

        Supported options are:

          -o, --overwrite  When a file already exists at the target
                           location, it is overwritten.
        """

        try:
            fi = self._dbfs.get_status(target)
            if fi.is_dir():
                target = os.path.join(target, os.path.basename(src))
        except: pass

        with progressbar.DataTransferBar() as bar:
            def update_cb(size, bytes_copied):
                bar.max_size=size
                bar.update(bytes_copied)

            self._dbfs.put(src, target, overwrite=overwrite, update_cb=update_cb)

    def _get_to_temp(self, src,  update_cb=None, prefix=".tmp-", suffix=None, **kwargs):
        try:
            if suffix is None:
                bn = posixpath.basename(src)
                try: suffix = bn[bn.rindex("."):]
                except: suffix = ""

            (f, target) = tempfile.mkstemp(prefix=prefix, suffix=suffix, **kwargs)
            out = os.fdopen(f, "wb")
            self._dbfs.get_to_file(src, out, update_cb=update_cb)
            out.close()
            return target

        except Exception as ex:
            try: out.close()
            except: pass
            try: os.remove(tmp_fn)
            except: pass
            raise ex

    @flag("overwrite", "o")
    @remote("src")
    @local("target", arity="?", default=".")
    def do_get(self, overwrite, src, target):
        """
        get [OPTS] src [target]

        Copies the given remote file to the local system.

        Supported options are as follows:

          -o, --overwrite  When a file already exists at the target
                           location, it is overwritten.
        """

        if os.path.isdir(target):
            target = os.path.join(target, os.path.basename(src))
        if os.path.exists(target) and not overwrite:
            raise Exception("Target file already exists")

        parent_dir = os.path.dirname(target)
        fastdbfs.util.mkdirs(parent_dir)

        with progressbar.DataTransferBar() as bar:
            def update_cb(size, bytes_copied):
                bar.max_size=size
                bar.update(bytes_copied)

            tmp_target = self._get_to_temp(src, update_cb=update_cb,
                                           prefix=".transferring-", suffix="", dir=parent_dir)
        os.rename(tmp_target, target)

    def _get_and_call(self, src, cb):
        target = self._get_to_temp(src)
        try:
            cb(target)
        finally:
            try: os.remove(target)
            except: pass

    @flag("recursive", "R")
    @remote("path")
    def do_rm(self, recursive, path):
        """
        rm [OPTS] path

        Supported options are as follows:

          -R, --recursive Delete files and directories recursively.
        """
        self._dbfs.rm(path, recursive=recursive)

    def _wrap_external_filter(self, cmd):
        if cmd is None:
            return None

        def filter(fis):
            data = { relpath: fi.to_data() for relpath, fi in fis.items() }
            return fastdbfs.util.call_external_processor_json(cmd, data)
        return filter

    @flag("quiet", "nowarn", "q")
    @flag("long", "ls", "l")
    @flag("human", "h")
    @_find_predicates
    @remote("path", default=".")
    def do_find(self, path, quiet, long, human, external_filter, **predicates):
        """
        find [OPTS] [RULES] [path]

        List files recursively according to a set of rules.

        Supported options are as follows:

          --nowarn     Do not report errors happening while
                       walking the file system tree.
          -l, --ls     Display file and directory properties.
          -h, --human  Show sizes like 1K, 210M, 4G, etc.

        In addition to those, find also supports the following set of
        options for selectively picking the files that should be
        displayed:

          --min-size=SIZE, --max-size=SIZE
                       Picks files according to their size (this rule
                       is ignored for directories).
          --newer-than=date, --older-than=date
                       Picks entries according to their modification time.
          --name=GLOB_PATTERN
                       Picks only files and directories whose basename
                       matches the given glob pattern.
          --iname=GLOB_PATTERN
                       Case insensitive version of "name".
          --re=REGULAR_EXPRESSION
                       Picks only entries whose basename matches the
                       given regular expression.
          --ire=REGULAR_EXPRESSION
                       Case insensitive version of "ire".
          --wholere=REGULAR_EXPRESSION
                       Picks file names whose relative path matches the
                       given regular expression.
          --iwholere=REGULAR_EXPRESSION
                       Case insensitive version of "wholere".
          --external-filter=CMD
                       Filters entries using an external command.

        Also, any of the rules above can be negated preceding it by
        "--exclude", for instance "--exclude-iname=*.jpeg"

        The rules above never cut the file system traversal. So for
        instance, a rule discarding some subdirectory, doesn't
        preclude that subdirectory for being traversed and its child
        entries picked.

        The basename is the last component of a patch. Relative paths
        are considered relative to the root directory passed as an
        argument.

        Example:

          find --ls --newer-than=yesterday --max-size=100K --iname=*.jpg
        """

        with progressbar.ProgressBar(redirect_stdout=True, redirect_stderr=True) as bar:
            def update_cb(entry, max_entries, done):
                bar.max_value = max_entries
                bar.update(done)
                if entry.good:
                    fi = entry.fi
                    relpath = fi.relpath(self._dbfs.cwd, path)
                    if long:
                        size = fi.size()
                        if human:
                            size= format_human_size(size)
                        mtime = format_time(fi.mtime())
                        print("{:>4} {:>12} {:>19} {}".format(fi.type(), size, mtime, relpath))
                    else:
                        print(relpath)
                if entry.ex and not quiet:
                    print("# Unable to recurse into {relpath}, {entry.ex}", file=sys.stderr)

            self._dbfs.find(path, update_cb,
                            filter_cb = self._wrap_external_filter(external_filter),
                            predicates = predicates)

    def _rput_update_cb(self, path, ok, ex):
        if ok:
            print(f"{path} ok!")
        elif ex is None:
            print(f"{path} FAILED!")
        else:
            print(f"{path} FAILED {ex}!")

    @flag("overwrite", "o")
    @local("src")
    @remote("target", arity="?")
    def do_rput(self, overwrite, src, target):
        """
        rput [src [target]]

        Copies the given local directory to the remote system
        recursively.
        """
        if target is None:
            normalized_src = os.path.normpath(src)
            # FIXME: this normalization is not gona work for Windows
            if normalized_src == "." or normalized_src == "/":
                target = "."
            else:
                target = os.path.basename(normalized_src)
        self._dbfs.rput(src, target, overwrite=overwrite, update_cb=self._rput_update_cb)

    @flag("verbose", "v")
    @flag("nowarn", "quiet", "q")
    @remote("src")
    @local("target", arity="?")
    @_find_predicates
    def do_rget(self, verbose, nowarn, src, target, external_filter, **predicates):
        """
        rget [OPTS] [RULES] [src [target]]

        Copies the given remote directory to the local system
        recursively.

        The options supported are as follows:

        -v, --verbose  Display the names of the files being copied.
        --nowarn       Do not show warnings.

        In addition to those, "rget" also accepts the same set of
        predicates as "find" for selecting the entries to copy.

        Example:
          rget --iname *.png --exclude-iwholere=/temp/ . /tmp/pngs
        """

        if verbose and nowarn:
            raise Exception("verbose and nowarn can not be used together")

        if target is None:
            # FIXME: Windows again
            normalized_src = os.path.normpath(src)
            if normalized_src == "." or normalized_src == "/":
                target = "."
            else:
                target = os.path.basename(normalized_src)

        with progressbar.ProgressBar(redirect_stderr=True) as bar:
            def update_cb(entry, max_entries, done):
                if not nowarn:
                    relpath = entry.fi.relpath(self._dbfs.cwd, src)
                    if entry.ex:
                        print(f"{relpath}: FAILED. {entry.ex}", file=sys.stderr)
                    elif entry.good and verbose:
                        print(f"{relpath}: copied.", file=sys.stderr)
                bar.max_value = max_entries
                bar.update(done)
            self._dbfs.rget(src, target,
                            update_cb,
                            filter_cb = self._wrap_external_filter(external_filter),
                            predicates = predicates)

    @remote("path")
    def do_cat(self, path):
        """
        cat path

        Prints the contents of the remote file.
        """
        def cb(fn):
            subprocess.run(["cat", "--", fn])
        self._get_and_call(path, cb)

    def _do_show(self, path, pager=None):
        if pager is None:
            pager = self.cfg("fastdbfs", "pager")
        def cb(fn):
            subprocess.run([pager, "--", fn])
        # print(f"pager: {pager}, cb: {cb}")
        self._get_and_call(arg, cb)

    @option("pager")
    @remote("path")
    def do_show(self, pager, path):
        """
        show path

        Display the contents of the remote file using your favorite
        pager.
        """

        # print(f"pager: {pager}, path: {path}")
        self._do_show(path, pager=pager)

    @remote("path")
    def do_more(self, path):
        self._do_show(path, pager="more")

    @remote("path")
    def do_less(self, path):
        self._do_show(path, pager="less")

    @remote("path")
    def do_batcat(self, path):
        self._do_show(path, pager="batcat")

    def _do_edit(self, path, editor):
        if editor is None:
            editor = self.cfg("fastdbfs", "editor", default=os.environ.get("EDITOR", "vi"))
        def cb(tmp_fn):
            # in order to avoid race conditions we force the mtime
            # of the temporal file into the past
            the_past = int(time.time() - 2)
            os.utime(tmp_fn, times=(the_past, the_past))
            subprocess.run([editor, "--", tmp_fn])
            stat_after = os.stat(tmp_fn)
            if (stat_after.st_mtime > the_past):
                with open(tmp_fn, "rb") as infile:
                    self._dbfs.put_from_file(infile, src, size=stat_after.st_size, overwrite=True)
            else:
                raise Exception(f"File was not modified!")

        self._get_and_call(path, cb)

    @option("editor")
    @remote("path")
    def do_edit(self, editor, path):
        """
        edit path

        Retrieves the remote file and opens it using your favorite editor.

        Once you closes the editor it copies the file back to the
        remote system.
        """
        self._do_edit(path, editor)

    @remote("path")
    def do_mg(self, arg):
        self._do_edit(path, editor="mg")

    @remote("path")
    def do_vi(self, arg):
        self._do_edit(path, editor="vi")

    def do_shell(self, arg):
        """
        !cmd args...

        Runs the given command locally.
        """
        os.system(arg)

    @argless()
    def do_EOF(self):
        print("\nBye!")
        return True

    @argless()
    def do_exit(self):
        """
        exit

        Exits the program.
        """

        print("Bye!")
        return True

    @argless()
    def do_q(self):
        print("Bye!")
        return True

    def _set_prompt(self):
        if self._dbfs:
            self.prompt = self._dbfs.prompt()
        else:
            self.prompt = "*disconnected* "

    def postcmd(self, stop, line):
        self._set_prompt()
        return stop

    def _tell_error(self, msg):
        _, ex, _ = sys.exc_info()
        print(f"{msg}: {type(ex).__name__} - {ex}")
        logging.debug("Stack trace", exc_info=True)

    def onecmd(self, line):
        try:
            return super().onecmd(line)
        except Exception as ex:
            self._tell_error("Operation failed")
            return False

    def preloop(self):
        # DEFAULTS contaminates everything, so we have to explicitly
        # estate the fields that we want passed to
        # logging.basicConfig().
        section = self._cfg["logging"]
        logcfg = { k: section[k]
                   for k in ("filename", "filemode", "format", "datefmt",
                             "style", "level", "encoding", "errors")
                   if k in section }

        print(f"logcfg: {logcfg}")

        logging.basicConfig(force=1, **logcfg)

        if "token" in self._cfg["DEFAULT"]:
            self._do_open("DEFAULT")
        self._set_prompt()

    def _parse(self, arg, *defaults, min=0, max=None):
        args = shlex.split(arg)
        max1 = max if max is not None else min + len(defaults)
        if ((len(args) < min) or
            (max1 >= 0 and len(args) > max1)):
            raise Exception("wrong number of arguments")

        args += defaults[len(args)-min:]

        return args



import shlex
import re
import os.path
import posixpath
import logging
import humanfriendly
import dateparser
import fnmatch

def _wrap(func):
    if hasattr(func, "arg_decls"):
        return func

    func_name = func.__name__

    def wrapper(self, cmd_line):
        kwargs = _parse_args(wrapper, cmd_line)
        logging.debug(f"Calling function {func_name} with arguments {kwargs}")
        return func(self, **kwargs)

    wrapper.__name__=func_name
    wrapper.__doc__=func.__doc__
    wrapper.arg_decls=[]
    return wrapper

def _normalize_key(name):
    return re.sub(r'\W', '_', name)

class _ArgDecl:
    def __init__(self, option=False, arity="1", default=None, preprocess=None,
                 cast=None, cast_args={}, **kwargs):
        self.option=option
        if option:
            self.names=kwargs["names"]
            self.name=self.names[0]
        else:
            self.name=kwargs["name"]
            self.names = [self.name]
        self.key=_normalize_key(self.name)
        self.arity=arity
        self.default=default
        self.cast=cast
        self.cast_args=cast_args
        self.preprocess=preprocess
        self.kwargs=kwargs

    def __str__(self):
        return f"ArgDecl({self.key}, names: {self.names}, option: {self.option}, arity: {self.arity}, default: {self.default})"

    def __repr__(self):
        return self.__str__()

def _add_arg_decl(func, **kwargs):
    wrapper = _wrap(func)
    arg_decl = _ArgDecl(**kwargs)
    wrapper.arg_decls.insert(0, arg_decl)
    return wrapper

def _option_lookup(option_decls, name):
    for decl in option_decls:
        for decl_name in decl.names:
            if decl_name == name:
                return decl
    raise Exception(f"Invalid option {name}")

def _compile_regex(str, case_insensitive):
    logging.debug(f"compiling pattern {str}, case_insensitive: {case_insensitive}")
    return re.compile(str, 0 if case_insensitive else re.I)

def _cast(arg, type, extra):
    # print(f"converting {arg} to {type}")
    if type == "int":
        return int(arg)
    if type == "size":
        return humanfriendly.parse_size(arg)
    if type == "glob":
        ci = extra.get("case_insensitive", False)
        return _compile_regex(fnmatch.translate(arg), ci)
    if type == "re":
        ci = extra.get("case_insensitive", False)
        return _compile_regex(arg, ci)
    if type.startswith("date"):
        settings = {"PREFER_DATES_FROM": "past"}
        if type == "date>":
            pass
        elif type == "date<":
            pass
        elif type == "date":
            pass
        else:
            raise Exception(f"Internal error: Invalid cast type {type}")
        dt = dateparser.parse(arg, settings=settings)
        # print(f"time limit for {type} {arg} is {dt} ({dt.timestamp()})")
        return dt.timestamp()
    if type == "date>":
        dt
    raise Exception(f"Internal error: Invalid cast type {type}")

def _parse_args(wrapper, cmdline):
    option_decls = []
    list_decls = []

    for arg_decl in wrapper.arg_decls:
        # print(arg_decl)
        if arg_decl.option:
            option_decls.append(arg_decl)
        else:
            list_decls.append(arg_decl)

    args = shlex.split(cmdline)
    kwargs = {}
    no_more_options = False
    while args:
        current = args.pop(0)
        if current.startswith("-") and not no_more_options:
            value = None
            if current == "--" or current == "-":
                no_more_options=True
                continue

            if current.startswith("--"):
                single_dash = False
                try:
                    offset=current[3:].index("=")+3
                    name=current[2:offset]
                    value=current[(offset+1):]
                except:
                    name=current[2:]
            else:
                single_dash = True
                name = current[1]
                if len(current) > 2:
                    value = current[2:]
            decl = _option_lookup(option_decls, name)
            if decl.arity == "0":
                kwargs[decl.key] = True
                if value is not None:
                    if single_dash:
                        # we allow the user to pack several options
                        # into one. For instance:
                        #     -lptfoo ==> -l -p -tfoo
                        args.insert(0, "-" + value)
                    else:
                        raise Exception(f"Option {name} doesn't take a value")
            else:
                if value is None:
                    if not args:
                        raise Exception(f"Option {name} requires a value")
                    value = args.pop(0)

                if decl.arity in ("1", "?"):
                    # we unwrap single values below.
                    kwargs[decl.key] = [value]
                elif decl.arity in ("*", "+"):
                    kwargs.setdefault(decl.key, []).append(value)
                else:
                    raise Exception(f"Internal error: unexpected arity {decl.arity} for {decl.key}")

        else: # so, it is not an option...
            for decl in list_decls:
                if decl.arity in ("*", "+"):
                    kwargs.setdefault(decl.key, []).append(current)
                    break
                if decl.key not in kwargs:
                    # we push an array anyway, even when the arity is
                    # 1 or ? in order to simplify the stealing logic
                    # below...

                    kwargs[decl.key]=[current]
                    break
            else:
                raise Exception(f"Too many arguments")

    # Steal from previous arguments in order to comply with arity
    # restrictions. We bubble up the arguments here!
    done = False
    while not done:
        done = True
        src = None
        surplus = 0
        for decl in list_decls:
            args = kwargs.get(decl.key, [])
            if args:
                if len(args) > 1:
                    src = decl
                    surplus = len(args) - 1
                else:
                    src = decl
            elif surplus > 0:
                kwargs[decl.key] = [kwargs[src.key].pop()]
                src = decl
                surplus -= 1
                done = False
            else:
                break

    # Now, we set the defaults, check arities and unwrap one value arguments and preprocess
    for decl in wrapper.arg_decls:
        args = kwargs.get(decl.key, None)
        if decl.arity in ("1", "?"):
            arg = args[0] if args else decl.default
            if decl.arity == "1" and arg is None:
                raise Exception(f"Missing mandatory argument {decl.name}")
            if decl.cast and arg is not None:
                arg = _cast(arg, decl.cast, decl.cast_args)
            if decl.preprocess:
                arg = decl.preprocess(arg)
            kwargs[decl.key] = arg
        else:
            if not args:
                args = decl.default
                if not args and decl.arity == "+":
                    raise Exception(f"Missing mandatory argument {decl.name}")
                if decl.preprocess:
                    args = [decl.preprocess(arg) for arg in args]
                kwargs[decl.key] = args

    return kwargs


# And now, finally, the decorators!!!

def option(*names, arity="?", **kwargs):
    def decorator(f):
        return _add_arg_decl(f, names=names, option=True, arity=arity, **kwargs)
    return decorator

def flag(*names, default=False, **kwargs):
    def decorator(f):
        return _add_arg_decl(f, names=names, option=True, arity="0", default=default, **kwargs)
    return decorator

def arg(name, arity="1", **kwargs):
    def decorator(f):
        return _add_arg_decl(f, name=name, arity=arity, **kwargs)
    return decorator

def remote(name, arity="1", **kwargs):
    def decorator(f):
        return _add_arg_decl(f, name=name, arity=arity, **kwargs)
    return decorator

def _expanduser(path):
    if path is None:
        return None
    return os.path.expanduser(path)

def local(name, arity="1", **kwargs):
    def decorator(f):
        return _add_arg_decl(f, name=name, arity=arity,
                             preprocess=_expanduser,
                             **kwargs)
    return decorator

def argless():
    def decorator(f):
        wrapper = _wrap(f)
        return wrapper
    return decorator

def chain(*decorators):
    def decorator(f):
        for d in decorators:
            f = d(f)
        return f
    return decorator




import asyncio
import aiohttp
import logging

seq = 0

class AlwaysComparable():
    def __init__(self, v):
        self.v = v

    def __lt__(a, b):
        av = a.v
        bv = b.v
        if bv is None: return True
        if av is None: return False
        if type(av) == type(bv): return av < bv
        if isinstance(av, int): return True
        if isinstance(bv, int): return False
        raise Exception("Internal error: don't know how to compare {type(av).__name__} with {type(bv).__name__}")

class Swarm:
    def __init__(self, max_workers, queue_max_size, use_priority_queue=False, name=None):
        global seq
        seq += 1
        self.name = name or f"swarm{seq}"
        queue_class = asyncio.PriorityQueue if use_priority_queue else asyncio.Queue
        self.queue = queue_class(maxsize = queue_max_size)
        self.max_workers = max_workers
        self.task_seq = 0


    async def run(self):
        logging.debug(f"Swarm {self.name} running")
        workers = [self._worker(ix) for ix in range(self.max_workers)]
        await asyncio.gather(*workers)

    async def run_while(self, task):
        logging.debug("Swarm {self.name} running")

        r = None
        ex = None

        async def wrapper():
            nonlocal r, ex
            try:
                r = await task
                #r = await asyncio.gather(task)[0]
            except Exception as ex1:
                logging.debug(f"Swarm {self.name} concurrent task failed, exception: {ex1}")
                ex = ex1
            await self.terminate()

        workers = [self._worker(ix) for ix in range(self.max_workers)]
        await asyncio.gather(wrapper(), *workers)

        if ex:
            logging.debug(f"Swarm {self.name} rethrowing exception {ex}")
            raise ex
        logging.debug(f"Swarm {self.name} run_while done")
        return r

    def loop_until_complete(self, task):
        loop = asyncio.get_event_loop()
        r = loop.run_until_complete(self.run_while(task))
        logging.debug(f"Swarm {self.name} loop_until_complete done")
        return r

    async def terminate(self):
        logging.debug(f"Swarm {self.name} terminating")
        for ix in range(self.max_workers):
            await self.put(None, task_key=ix)

    async def _worker(self, ix):
        logging.debug(f"Worker {ix} started")

        async with aiohttp.ClientSession() as session:
            while True:
                logging.debug(f"Swarm {self.name}/{ix} waiting for tasks")
                (task_key, cb, response_queue, kwargs) = await self.queue.get()
                task_key = task_key.v
                if cb is None:
                    logging.debug(f"Swarm {self.name}/{ix} terminating")
                    return
                try:
                    logging.debug(f"Swarm {self.name}/{ix} starting task {task_key}")
                    value = await cb(session, **kwargs)
                    res = (task_key, value, None)
                except Exception as ex:
                    logging.debug(f"Swarm {self.name}/{ix} running task {task_key} failed with exception {ex}")
                    res = (task_key, None, ex)

                logging.debug(f"Swarm {self.name}/{ix} ended task {task_key}")
                if response_queue:
                    await response_queue.put(res)
                    logging.debug(f"Swarm {self.name}/{ix} queued response for task {task_key}")

    async def put(self, task, task_key=None, response_queue=None, **kwargs):
        if task_key is None:
            task_key = self.task_seq
            self.task_seq += 1
        wrapper = (AlwaysComparable(task_key), task, response_queue, kwargs)
        logging.debug(f"Swarm {self.name} queueing task {task and task.__name__} with key {task_key}")
        return await self.queue.put(wrapper)

    @staticmethod
    def unwrap_response(res):
        (_, value, ex) = res
        if ex is not None:
            logging.debug(f"Rethrowing exception received from Swarm: {ex}")
            raise ex
        return value
